require 'mscorlib'
require 'System.Windows.Forms'
require 'System.Drawing'
require 'OpenTK'
require 'OpenTK.GLControl'
include System::Drawing
include OpenTK
include OpenTK::Graphics
include OpenTK::Graphics::OpenGL


class MyGLPanelBase 
  def initialize(form_panel)
    @loaded = false
    #
    @panel = form_panel
    #
    @glview = OpenTK::GLControl.new()
    @glview.Load.add( proc { eOtkLoad() } )
    @glview.Resize.add( proc { eOtkResize } )
    @glview.Paint.add( proc { eOtkPaint } )
    @glview.MouseDown.add( proc { |sender,e| eOtkMouseButtonDown(sender, e) } )
    @glview.MouseUp.add( proc { |sender,e| eOtkMouseButtonUp(sender, e) } )
    @glview.MouseMove.add( proc { eOtkMouseMove } )
    @glview.MouseWheel.add( proc { |sender,e| eOtkMouseWheel(sender, e) } )
    @glview.Dock = System::Windows::Forms::DockStyle.Fill
    @panel.Controls.add(@glview)
    #
    @eye_roll_x = 0
    @eye_roll_y = 0
    @eye_roll_z = 0
    @eye_distance = 3.0
    @eye_target_x = 0
    @eye_target_y = 0
    @eye_target_z = 0
    #
    @mouse_drag = false
    @mouse_pos_x = 0
    @mouse_pos_y = 0
  end

  attr_accessor :glview

  # --- GLControl event

  def eOtkLoad
    GL.clear_color(Color.dark_blue)
    GL.enable(EnableCap.depth_test)
    @loaded = true
    refresh
  end
  
  def eOtkResize
    GL.viewport(0, 0, @glview.width, @glview.height)
    aspect_ratio = @glview.width.to_f / @glview.height.to_f

    perspective = OpenTK::Matrix4.create_perspective_field_of_view(
      MathHelper.PiOver4,
      aspect_ratio,
      0.1, 1000.0)
    GL.matrix_mode(MatrixMode.projection)
    GL.load_matrix(perspective)
    refresh
  end
  
  def eOtkPaint
    return if @loaded == false
    #
    refresh
  end
  
  # --- mouse event
  
  def eOtkMouseButtonDown(sender, e)
    @mouse_drag = true
    mouse_down(e)
  end
  
  def eOtkMouseButtonUp(sender, e)
    if @mouse_drag == true then
      refresh
      @mouse_drag = false
    end
    mouse_up(e)
  end

  def eOtkMouseMove
    mouse_move
  end

  def eOtkMouseWheel(sender, e)
    mouse_change_wheel(e)
  end

  # --- view point

  def camera(x, y, z)
    @eye_roll_x = x
    @eye_roll_y = y
    @eye_roll_z = z
  end

  def deg2rad( deg )
    rad = Math::PI * 2 * deg / 360.0
  end

  def cal_lookat
    eye_x = 0
    eye_y = 0
    eye_z = @eye_distance
    rx = deg2rad(@eye_roll_x)
    ry = deg2rad(@eye_roll_y)
    rz = deg2rad(@eye_roll_z)
    # - roll x
    tx = eye_x
    ty = eye_y*Math.cos(rx) + eye_z*Math.sin(rx)
    tz = -eye_y*Math.sin(rx) + eye_z*Math.cos(rx)
    eye_x = tx
    eye_y = ty
    eye_z = tz
    # - roll y
    tx = eye_x*Math.cos(ry) - eye_z*Math.sin(ry)
    ty = eye_y
    tz = eye_x*Math.sin(ry) + eye_z*Math.cos(ry)
    eye_x = tx
    eye_y = ty
    eye_z = tz
    # - roll z
    #tx = eye_x*Math.cos(rz) + eye_y*Math.sin(rz)
    #ty = -eye_x*Math.sin(rz) + eye_y*Math.cos(rz)
    #tz = eye_z
    #eye_x = tx
    #eye_y = ty
    #eye_z = tz
    eye_x += @eye_target_x
    eye_y += @eye_target_y
    #
    frame_roll_x = Math.sin( rz )
    frame_roll_y = Math.cos( rz )
    lookat = Matrix4.look_at(eye_x, eye_y, eye_z,
                             @eye_target_x, @eye_target_y, @eye_target_z,
                             frame_roll_x,  frame_roll_y, 0)
  end

  
  # -----------------------------------------------------
  #  draw frame
  # -----------------------------------------------------

  def refresh
    @glview.MakeCurrent()

    GL.Clear(ClearBufferMask.ColorBufferBit | ClearBufferMask.DepthBufferBit);
    GL.MatrixMode(MatrixMode.Modelview);
    GL.load_matrix( cal_lookat() )

    draw()

    @glview.swap_buffers
  end

  def draw
    # -- inherited
  end  


  # -----------------------------------------------------
  #  mouse event
  # -----------------------------------------------------

  def mouse_down(e)
    pos = @panel.PointToClient(System::Windows::Forms::Control.MousePosition)
    @mouse_pos_x = pos.X
    @mouse_pos_y = pos.Y
  end

  def mouse_up(e)
  end

  def mouse_move
    return if @mouse_drag == false
    pos = @panel.PointToClient(System::Windows::Forms::Control.MousePosition)
    mx = pos.X
    my = pos.Y
    dx = (@mouse_pos_x - mx) / 2.0
    dy = (@mouse_pos_y - my) / 2.0
    @eye_roll_x = (@eye_roll_x - dy) % 360
    @eye_roll_y = (@eye_roll_y - dx) % 360
    @mouse_pos_x = mx
    @mouse_pos_y = my
    @refresh = true
    refresh
  end

  def mouse_change_wheel(e)
    wd = e.Delta / 120.0
    @eye_distance += (wd / 2.0)
    if @eye_distance < 0.1 then
      @eye_distance = 0.1
    end
    if @eye_distance > 1000.0 then
      @eye_distance = 1000.0
    end
    @mouse_wheel = move
    @refresh = true
  end

end

