require "pattern_canvas.rb"

# --------------------------------------------------------
#  plot work
# --------------------------------------------------------

class PatternPlotter
  include LineCrossCheck

  def initialize
    @mil_size = 0.0
    @pad_size = 0.5
    @flash_pattern = 0
    @pattern = []
    @circle_resolution = 12
    @marge_outline = true
  end

  attr_accessor :circle_resolution
  attr_accessor :marge_outline

  def clear_pattern
    @pattern.clear
  end

  def set_mil_size(size)
    @mil_size = size
  end

  def set_pad_size(size)
    @pad_size = size
  end

  def set_pad_round(type)
    if type == true then
      # round pattern
      @flash_pattern = 1
    else
      # square pattern
      @flash_pattern = 0
    end
  end

  def set_round_resolution(div)
    @circle_resolution = div
  end

  def get_pattern_num
    @pattern.size
  end

  # -----------------------------
  # --- do plot

  def flash(x, y)
    item = PatternObject.new
    item.circle_resolution = @circle_resolution
    item.set_mil_size(@mil_size)
    item.set_pad_size(@pad_size)
    item.set_pad_round(@flash_pattern == 1)
    item.flash(x, y)
    add_item(item)
  end

  def move(x1, y1, x2, y2)
    item = PatternObject.new
    item.circle_resolution = @circle_resolution
    item.set_mil_size(@mil_size)
    item.set_pad_size(@pad_size)
    item.set_pad_round(@flash_pattern == 1)
    item.move(x1, y1, x2, y2)
    add_item(item)
  end

  def draw_polygon(poly)
    # polygon
    item = PatternObject.new
    item.start_polygon
    poly.each do |line|
      item.add_polyline(line[0].x, line[0].y, line[1].x, line[1].y)
    end
    item.end_polygon
    add_item(item)
    # hatch
    min_x = max_x = poly[0].get_point(0).x
    min_y = max_y = poly[0].get_point(0).y
    poly.each do |line|
      min_x=line[0].x if min_x>line[0].x
      min_x=line[1].x if min_x>line[1].x
      min_y=line[0].y if min_y>line[0].y
      min_y=line[1].y if min_y>line[1].y
      max_x=line[0].x if max_x<line[0].x
      max_x=line[1].x if max_x<line[1].x
      max_y=line[0].y if max_y<line[0].y
      max_y=line[1].y if max_y<line[1].y
    end
    scan_y = min_y - 0.05
    while(scan_y < max_y) do
      scan_y += @pad_size * 0.95
      scan_y = max_y if scan_y > max_y
      
      point = scanline_across(poly, scan_y, min_x-1.0, max_x+1.0)
      next if (point.size % 2) == 1
      point.sort! do |a, b|
        (a.x - b.x)
      end
      for i in 0...(point.size / 2)
        p1 = point[i*2].x + (@pad_size/2.0)
        p2 = point[i*2+1].x - (@pad_size/2.0)
        next if (p2 - p1) < 0.0
        item = PatternObject.new
        item.circle_resolution = 8
        item.set_mil_size(@mil_size)
        item.set_pad_size(@pad_size)
        item.set_pad_round(true)
        item.move(p1, scan_y, p2, scan_y)
        add_item(item)
      end
    end
    # outline
    poly.each do |line|
      item = PatternObject.new
      item.circle_resolution = @circle_resolution
      item.set_mil_size(@mil_size)
      item.set_pad_size(@pad_size)
      item.set_pad_round(true)
      item.move(line[0].x, line[0].y, line[1].x, line[1].y)
      add_item(item)
    end
  end

  def scanline_across(poly, raster_y, min_x, max_x)
    point_list = []
    raster = PLine.new
    raster.set_point(0, PPoint.new(min_x, raster_y))
    raster.set_point(1, PPoint.new(max_x, raster_y))
    poly.each do |line|
      if cross?(line, raster) then
        point_list << crosspoint(line, raster)
      end
    end
    return point_list
  end


  def add_item(item)
    @pattern << item
  end

  def prepare_outline(thread_num)
    tlist = []
    for num in 0...thread_num
      size = (@pattern.size / thread_num) + 1
      startnum = size * num
      endnum = startnum + size
      if endnum > @pattern.size then
        endnum = @pattern.size
      end
      t = Thread.new(startnum, endnum) do |i_from, i_to|
        for loop in i_from...i_to
          item = @pattern[loop]
          item.drip_mil_path
        end
      end
      tlist << t
    end
    tlist.each do |t|
      t.join
    end
  end

  def extruct_outline(num)
    item = @pattern[num]
    lines = item.get_mil_path
    deflines = []
    for npat in 0...@pattern.size
      next if npat == num
      prim = @pattern[npat]
      if item.collision_bounding_box?(prim.bounding_box) then
        if npat < num then
          newlines = prim.marge_outline(lines, false)
        else
          newlines = prim.marge_outline(lines, true)
        end
        item.set_mil_path(newlines)
        lines = newlines
      end
    end
  end


  # -----------------------------
  # --- result

  def get_object
    newobj = []
    @pattern.each do |prim|
      prim.get_pattern_object.each do |tri|
        newobj << tri
      end
    end
    return newobj
  end

  def get_outline
    newobj = []
    @pattern.each do |prim|
      prim.get_mil_path.each do |line|
        newobj << line
      end
    end
    return newobj
  end

  def get_outline_bound(num)
    newobj = []
    if num > @pattern.size then
      num = @pattern.size
    end
    for index in 0...num
      prim = @pattern[index]
      prim.get_mil_path.each do |line|
        newobj << line
      end
    end
    return newobj
  end

end
